#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import click, os, time, fnmatch, json
import re
import sys
import SocketServer
import SimpleHTTPServer, BaseHTTPServer
import sass, jsmin, coffeescript, csscompressor, htmlmin, markdown
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from jsonschema import validate, exceptions
from cStringIO import StringIO

default = {
    "config" : {
        "output_suffix" : "build",
        },
    "jobs" : [
        {
            "type"     : "javascript",
            "watch"    : ["*.js"],
            "compress" : True,
        },
        {
            "type"   : "sass",
            "watch"  : ["*.scss","*.sass"],
            "compress" : True,
        },
        {
            "type"   : "coffeescript",
            "watch"  : ["*.coffee"],
            "compress" : True,
        },
        {
            "type"   : "css",
            "watch"  : ["*.css"],
            "compress" : True,
        },
        {
            "type"   : "html",
            "watch"  : ["*.htm","*.html"],
            "compress" : True,
        },
        {
            "type"   : "markdown",
            "watch"  : ["*.md","*.markdown"],
            "compress" : True,
        },
        {
            "type"   : "plaintext",
            "watch"  : ["*.txt","*.text"],
        }
    ]  
}

job_schema = {
    "type" : "object",
    "properties" : {
        "type" : {"type" : "string"},
        "files" : {"type" : "array"},
        "watch" : {"type" : "array"},
    },
    "required" : ["type","watch"],
}

config_schema = {
    "type" : "object",
    "properties" : {
        "jobs" : { "type" : "array", "items" : job_schema }
    }
}


class Config(object):
    def __init__(self):
        self.config = None

pass_config = click.make_pass_decorator(Config, ensure=True)


class ThreadingSimpleServer(SocketServer.ThreadingMixIn,
                            BaseHTTPServer.HTTPServer):
    pass

 
class checkEvent(FileSystemEventHandler):
    def __init__(self, config, watch_list):
        self.config = config
        self.watch_list = watch_list
    def on_any_event(self, event):
        if event.event_type in ['modified','created']:
            for i in self.watch_list:
                if fnmatch.fnmatch(event.src_path, i) and '.%s.' % self.config['config']['output_suffix'] not in event.src_path:
                    click.echo(click.style('Change Detected',bold=True))
                    for j in self.watch_list[i]:
                        do_job(event.src_path, self.config['jobs'][j], self.config['config']['output_suffix'])


def build_watch_list(jobs):
    watch_list = {}
    for i in jobs:
        for f in i['watch']:
            if f in watch_list:
                if jobs.index(i) not in watch_list[f]:
                    watch_list[f].append( jobs.index(i) )
            else:
                watch_list[f] = [jobs.index(i)]

    return watch_list


def template(path, data, tlevel='  +'):
    regex = re.compile(r'\.\.\.inc.*\.\.\.',re.LOCALE) # HTML, Markdown
    for match in regex.findall(data):
        f = match.split(' ')[1].rstrip('.').strip()
        f = os.path.join(path,f)

        if os.path.exists(f) == False:
            click.echo(click.style(tlevel, fg='red')+'%s' % f)
        else: 
            click.echo(click.style(tlevel, fg='green')+'%s' % f)
            with open(f) as fo:
                child_data = fo.read()
                child_data = child_data.decode('UTF-8')
            child_data = template(os.path.dirname(f), child_data, '  '+tlevel)
            data = data.replace(match, child_data)
   
    return data


def do_job(fpath, job, output_suffix):
    """Simple Build Tool for Webdeveloper"""
    if 'files' not in job:
        files = [fpath] 
    else:
        files = job['files']

    for f in files:

        if output_suffix not in f:
            click.echo(click.style('==>', fg='green')+'%s' % f)
            with open(f) as fo:
                data = fo.read()
            data = data.decode('UTF-8')
            
            output = template(os.path.dirname(f), data)
            if job['type'] in ['sass', 'scss']:
                f = re.sub(r"\.scss|\.sass", r".%s.css" % output_suffix, f)            
                if job['compress'] == True:
                    output = sass.compile(string=output, output_style='compressed')
                else:
                    output = sass.compile(string=output, output_style='expanded')
            if job['type'] in ['coffeescript']:
                f = re.sub(r"\.coffee", r".%s.js" % output_suffix, f)            
                output = coffeescript.compile(output)
                if job['compress'] == True:
                    output = jsmin.jsmin(output)
            if job['type'] in ['markdown']:
                f = re.sub(r"\.md|\.markdown", r".%s.html" % output_suffix, f)            
                output = markdown.markdown(output)
                if job['compress'] == True:
                    output = htmlmin.minify(output)

            if job['type'] in ['css']:
                f = re.sub(r"\.(?=[^.]*$)", r".%s." % output_suffix, f) 
                if job['compress'] == True:
                    output = csscompressor.compress(output)
            if job['type'] in ['html']:
                f = re.sub(r"\.(?=[^.]*$)", r".%s." % output_suffix, f) 
                if job['compress'] == True:
                    output = htmlmin.minify(output)
            if job['type'] in ['javascript']:
                f = re.sub(r"\.(?=[^.]*$)", r".%s." % output_suffix, f) 
                if job['compress'] == True:
                    output = jsmin.jsmin(output)
            if job['type'] in ['plaintext']:
                f = re.sub(r"\.(?=[^.]*$)", r".%s." % output_suffix, f) 

            with open(f, 'w') as fh:
                fh.write(output.encode('UTF-8'))
            click.echo(click.style('<==', fg='blue')+'%s' % f)


@click.group()
@pass_config
def cli(config):
    '''Jidōka is a build tool for Web Developer. The name can described as 
    "intelligent automation" or "automation with a human touch."'''

    if os.path.exists('./jidoka.json'):
        with open('./jidoka.json') as f:
            try:
                config.config = json.loads(f.read())
            except ValueError as e:
                click.echo(click.style('Check your config: %s' % e.message, fg='red'))
                exit()

        try:
            validate(config.config, config_schema)
        except exceptions.ValidationError as e:
            click.echo(click.style('jidoka.json %s' % e.message, fg='red'))
            exit()
    else:
        config.config = default

    pass


@cli.command()
@pass_config
def init(config):
    """Saves a basic project configuration to jidoka.json"""
    if os.path.exists('./jidoka.json'):
        click.echo('This is already a jidoka project!')
    else:
        with open('./jidoka.json','w') as f:
            f.write(json.dumps(config.config, indent=4, sort_keys=True))
        click.echo('Initialise new project, modify jidoka.json to fit your requirements.')


@cli.command()
@pass_config
def serve(config):
    """Start a local HTTPserver"""
    click.echo(click.style('Start development server on http://127.0.0.1:8080', bold=True))
    port = 8000

    server = ThreadingSimpleServer(('', port), SimpleHTTPServer.SimpleHTTPRequestHandler)
    try:
        while True:
            #sys.stdout.flush()
            server.handle_request()
    except KeyboardInterrupt:
        pass


@cli.command()
@pass_config
def build(config):
    """Build your project"""
 
    watch_list = build_watch_list(config.config['jobs'])

    fileList = []
    for root, subFolders, files in os.walk('.'):
        for f in files:
            f = os.path.join(root,f)
            for i in watch_list:
                if fnmatch.fnmatch(f, i) and '.min.' not in f:
                    for j in watch_list[i]:
                        do_job(f, config.config['jobs'][j], config.config['config']['output_suffix'])


@cli.command()
@pass_config
def watch(config):
    """watch your project and rebuild files"""

    watch_list = build_watch_list(config.config['jobs'])

    event_handler = checkEvent(config.config, watch_list)
    observer = Observer()
    observer.schedule(event_handler, path='.', recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


if __name__ == '__main__':
    cli()
